// -*- C++ -*-
// $RCSfile: ooferror.h,v $
// $Revision: 1.7 $
// $Author: langer $
// $Date: 2004/06/04 20:57:26 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#ifndef OOFERROR_H
#define OOFERROR_H

#include <string>
#include <Python.h>

// Error namespace causes too many problems for SWIG.  Instead,
// use a customizing prefix for error names.
//

// Separate "stub" class for passing Python errors out through C++.
// This happens when a callback sets the Python error state, but 
// doesn't return to the Python environment.  The C++ caller should
// detect the error (by NULL return), and throw this exception,
// which will be caught in the exception typemap.
class PythonError {
public:
  PythonError() {}
};

// base class so that all OOF errors can be caught by the one catch
// statement typemap in oof.swg.  Also holds the One True Pythonizer
// for error classes.

class ErrError {
public:
  static PyObject *pyconverter;
  virtual ~ErrError() {}
  virtual std::string pythonequiv() const = 0;
};

void pyErrorInit(PyObject*);

class IndexError : public ErrError {
public:
  virtual std::string pythonequiv() const;
};

// Programming errors are fatal (to the program...)

class ErrProgrammingError : public ErrError {
private:
  const std::string file;
  const int line;
  const std::string msg;
public:
  ErrProgrammingError(const std::string &f, int l);
  ErrProgrammingError(const std::string &m, const std::string &f, int l);
  virtual ~ErrProgrammingError() {}
  virtual std::string message() const;
  std::string filename() const { return file; } // For reading from Python.
  int lineno() const { return line; }           // ... this one too.
  virtual std::string pythonequiv() const;
};


// Resource shortages are probably fatal.

class ErrResourceShortage : public ErrError {
private:
  const std::string msg;
public:
  ErrResourceShortage(const std::string &m) 
    : msg(m) 
  {}
  virtual ~ErrResourceShortage() {}
  virtual std::string message() const;
  virtual std::string pythonequiv() const;
};

class ErrBadIndex : public ErrProgrammingError {
private:
  int badindex;
public:
  ErrBadIndex(int i, const std::string &f, int l)
    : ErrProgrammingError(f, l), badindex(i)
  {}
  virtual std::string message() const;
  virtual std::string pythonequiv() const;
};


// User errors shouldn't be fatal.

class ErrUserError : public ErrError {
public:
  const std::string msg;  // Public so Python can see it.
  ErrUserError(const std::string &m) : msg(m) {}
  virtual ~ErrUserError() {}
  virtual std::string message() const { return msg; }
  virtual std::string pythonequiv() const;
};

class ErrSetupError : public ErrUserError {
public:
  ErrSetupError(const std::string &m)
    : ErrUserError(m)
  {}
  virtual ~ErrSetupError() {}
  virtual std::string pythonequiv() const;
};

class ErrNoCanvas : public ErrError {
public:
  virtual std::string pythonequiv() const;
  ErrNoCanvas();
  virtual ~ErrNoCanvas();
};

void testerror();

#endif
