// -*- C++ -*-
// $RCSfile: pixelgroup.C,v $
// $Revision: 1.30 $
// $Author: langer $
// $Date: 2004/07/14 16:26:37 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#include "common/IO/bitoverlay.h"
#include "common/array.h"
#include "common/cmicrostructure.h"
#include "common/pixelattribute.h"
#include "common/pixelgroup.h"
#include "common/switchboard.h"
#include "common/tostring.h"
#include "common/trace.h"
#include <algorithm>		// std::sort


// TODO LATER: Would it be more efficient to base PixelGroup on a
// std::set instead of a std::vector?

//----------- 

const std::string
PixelGroupAttributeRegistration::classname_("PixelGroupAttributeRegistration");
const std::string
PixelGroupAttributeRegistration::modulename_("SWIG.common.pixelgroup");

static PixelGroupAttributeRegistration *reg = 0;

PixelGroupAttributeRegistration::PixelGroupAttributeRegistration()
  : PxlAttributeRegistration("PixelGroups")
{
  reg = this;
}

// -----------

const std::string PixelGroup::classname_("PixelGroup");
const std::string PixelGroup::modulename_("SWIG.common.pixelgroup");

int PixelGroup::ngroups(0);
  
PixelGroup::PixelGroup(const std::string &name, const ICoord *geometry,
		       CMicrostructure *microstructure)
  : name_(name),
    weeded(true),
    geometry(*geometry),
    id_(ngroups++),
    suppress_announce(0),
    microstructure(microstructure),
    meshable_(1),
    defunct_(0)
{
}

PixelGroup::PixelGroup(const std::string &name, const PixelGroup &other)
  : name_(name),
    weeded(true),
    geometry(other.geometry),
    id_(ngroups++),
    microstructure(other.microstructure),
    meshable_(other.meshable_),
    defunct_(other.defunct_)
{
  add(other.members());
}

PixelGroup::~PixelGroup() {}

void PixelGroup::resize(const ICoord *newgeom) {
  if(geometry != *newgeom) {
    geometry = *newgeom;
    // Remove all pixels that don't fit in the new geometry.
    weed();
    int n = len();
    int j = 0;			// new index of accepted pixel
    for(int i=0; i<n; i++) {	// loop over old set of pixels
      ICoord pxl = members_[i];
      if(pxl(0) < geometry(0) && pxl(1) < geometry(1)) { // accept this pixel
	if(j < i)
	  members_[j] = members_[i];
	j++;
      }
    }
    members_.resize(j);
  }
}

void PixelGroup::set_defunct() {
  defunct_ = true;
  if(microstructure) {
    Array<PixelAttribute*> &groupMap = reg->map(microstructure);
    weed();
    for(int i=0; i<len(); i++) {
      GroupList *list = dynamic_cast<GroupList*>(groupMap[members_[i]]);
      list->resort();
    }
  }
}

void PixelGroup::add(const std::vector<ICoord> *pixels) {
  members_.reserve(members_.size() + pixels->size());
  for(int i=0; i<pixels->size(); i++)
    members_.push_back((*pixels)[i]);
  weeded = false;
  // Update grouplists in microstructure
  if(microstructure) {
    Array<PixelAttribute*> &groupMap = reg->map(microstructure);
    for(std::vector<ICoord>::const_iterator i=pixels->begin(); i<pixels->end();
	++i) {
      GroupList *list = dynamic_cast<GroupList*>(groupMap[*i]);
      list->add(this);
    }
    microstructure->recategorize();
  }
  groupchanged();
}

void PixelGroup::add(const ICoord &pixel) {
  members_.push_back(pixel);
  weeded = false;
  if(microstructure) {
    Array<PixelAttribute*> &groupMap = reg->map(microstructure);
    GroupList *list = dynamic_cast<GroupList*>(groupMap[pixel]);
    list->add(this);
    microstructure->recategorize();
  }
  groupchanged();
}

void PixelGroup::remove(const std::vector<ICoord> *pixels) {
  std::vector<ICoord> rejects(*pixels);	// copy, so we can sort
  weed();			// sort and weed the group
  std::sort(rejects.begin(), rejects.end()); // sort pixels to be rejected
  int g = 0;			// next group member to consider
  int r = 0;			// next reject to consider
  int newsize = 0;

  while(g < members_.size() && r < rejects.size()) {
    if(r < rejects.size()-1 && rejects[r]==rejects[r+1]) {
      r++;			// skip duplications in reject list
    }
    else if(members_[g] == rejects[r]) {
      g++;			// skip rejects in member list
    }
    else if(members_[g] < rejects[r]) {
      // retain nonrejected members
      if(g > newsize)
	members_[newsize] = members_[g];
      g++;
      newsize++;
    }
    else {
      // go on to next reject
      r++;
    }
  }
  // Include all remaining group members left after we've examined all
  // the rejects.
  while(g < members_.size()) {
    members_[newsize++] = members_[g++];
  }
  members_.resize(newsize);

  // Update grouplists in microstructure
  if(microstructure) {
    Array<PixelAttribute*> &groupMap = reg->map(microstructure);
    for(std::vector<ICoord>::const_iterator i=pixels->begin(); i<pixels->end();
	++i) {
      GroupList *list = dynamic_cast<GroupList*>(groupMap[*i]);
      list->remove(this);
    }
    microstructure->recategorize();
  }

  groupchanged();
}

ICoord PixelGroup::pop() {
  weed();
  ICoord pxl = members_[members_.size()-1];
  members_.pop_back();
  groupchanged();
  return pxl;
}

void PixelGroup::clear() {
  // Update grouplists in microstructure
  if(microstructure) {
    Array<PixelAttribute*> &groupMap = reg->map(microstructure);
    for(std::vector<ICoord>::const_iterator i=members_.begin(); i<members_.end();
	++i) {
      GroupList *list = dynamic_cast<GroupList*>(groupMap[*i]);
      list->remove(this);
    }
    microstructure->recategorize();
  }

  members_.clear();
  weeded = true;
  groupchanged();
}

// void PixelGroup::invert(const ICoord *geometry) {
//   weed();
//   Array<bool> arr(*geometry, true); // initialized to true
//   for(int i=0; i<members_.size(); i++)
//     arr[members_[i]] = false;
//   members_.clear();
//   for(Array<bool>::iterator ij=arr.begin(); ij!=arr.end(); ++ij)
//     if(*ij)
//       members_.push_back(ij.coord());
//   weeded = 0;			// because order might not be canonical
// }

void PixelGroup::setFromBitmap(const BitmapOverlay &bitmap) {
  const ICoord &bitmapsize = bitmap.sizeInPixels();
  int xmin = bitmapsize(0) < geometry(0) ? bitmapsize(0) : geometry(0);
  int ymin = bitmapsize(1) < geometry(1) ? bitmapsize(1) : geometry(1);
  const Array<bool> sub(bitmap.data.subarray(ICoord(0,0), ICoord(xmin, ymin)));
  for(Array<bool>::const_iterator i=sub.begin(); i!=sub.end(); ++i)
    if(*i)
      members_.push_back(i.coord());
  weeded = 0;
  groupchanged();
}

void PixelGroup::copy(const PixelGroup *other) {
  int n = other->len();
  members_.resize(n);
  const std::vector<ICoord> &otherpts = *other->members();
  for(int i=0; i<n; i++)
    members_[i] = otherpts[i];
}

int PixelGroup::len() const {
  weed();
  return members_.size();
}

void PixelGroup::weed() const {
  if(weeded) return;
  if(!members_.empty()) {
    std::sort(members_.begin(), members_.end());
    std::vector<ICoord>::iterator newend = std::unique(members_.begin(),
						       members_.end());
    members_.erase(newend, members_.end());
   }
  weeded = true;
}

void PixelGroup::groupchanged() const {
  if(!suppress_announce) {
    OOFMessage msg = OOFMessage("changed pixel group");
    msg.addarg(*this);
    switchboard_notify(msg);
  }
}

const std::vector<ICoord> *PixelGroup::members() const {
  weed();
  return &members_;
}

void PixelGroup::set_meshable(bool deshabille) {
  meshable_ = deshabille;
  groupchanged();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// The CMicrostructure keeps an array of GroupList objects, indicating
// to which groups each pixel belongs.  GroupList is a subclass of
// PixelAttribute, and as such is used to categorize pixels.

void GroupList::add(PixelGroup *group) {
  data.push_back(group);
  sorted = false;
}

void GroupList::remove(PixelGroup *group) {
  sort();
  for(std::vector<PixelGroup*>::iterator i=data.begin(); i<data.end(); ++i) {
    if(*i == group) {
      data.erase(i);
      return;
    }
  }
}

const std::vector<PixelGroup*> &GroupList::members() const {
  sort();
  return data;
}

void GroupList::sort() const {
  if(sorted) return;
  // remove defunct groups
  for(std::vector<PixelGroup*>::iterator i=data.begin(); i<data.end(); ++i) {
    if((*i)->is_defunct())
      i = data.erase(i);
  }
  // sort the remainder
  std::sort(data.begin(), data.end(), groupsorter);
  // remove duplicates
  std::vector<PixelGroup*>::iterator newend = 
    std::unique(data.begin(), data.end());
  data.erase(newend, data.end());
  sorted = true;
}

bool GroupList::groupsorter(const PixelGroup *a, const PixelGroup *b) {
  return a->name() < b->name();
}

bool GroupList::operator<(const PixelAttribute &other) const {
  // Compare two lists of groups, skipping those that aren't meshable.
  const GroupList &otherlist = dynamic_cast<const GroupList&>(other);
  const std::vector<PixelGroup*> &grps = members();
  const std::vector<PixelGroup*> &othergrps = otherlist.members();
  std::vector<PixelGroup*>::const_iterator i= grps.begin();
  std::vector<PixelGroup*>::const_iterator j = othergrps.begin();
  while(i<grps.end() && j<othergrps.end()) {
    if(!(*i)->is_meshable())
      ++i;
    else if(!(*j)->is_meshable())
      ++j;
    else if(*i < *j)
      return true;
    else if(*i > *j)
      return false;
    ++i;
    ++j;
  }
  if(i == grps.end() && j != othergrps.end())
    return true;
  return false;
}

void GroupList::print(std::ostream &os) const {
  const std::vector<PixelGroup*> &grps = members();
  if(grps.size() > 0) {
    os << "Groups(";
    for(int i=0; i<grps.size(); i++) {
      if(i > 0) os << ", ";
      os << grps[i]->name();
    }
    os << ")";
  }
  else
    os << "[No Groups]";
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

std::vector<std::string> *pixelGroupNames(const CMicrostructure *microstructure,
					  const ICoord *where)
{
  Array<PixelAttribute*> groupMap = reg->map(microstructure);
  GroupList *list = dynamic_cast<GroupList*>(groupMap[*where]);
  const std::vector<PixelGroup*> &groups = list->members();
  std::vector<std::string> *roster = new std::vector<std::string>;
  roster->reserve(groups.size());
  for(int i=0; i<groups.size(); i++)
    roster->push_back(groups[i]->name());
  return roster;
}
