# -*- python -*-
# $RCSfile: boundarycondition.py,v $
# $Revision: 1.49 $
# $Author: langer $
# $Date: 2004/09/02 18:52:42 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from SWIG.common import switchboard
from common import labeltree
from common import utils
from common.IO import automatic
from common.IO import oofmenu
from common.IO import parameter
from common.IO import whoville
from engine import bdycondition
from engine import skeletoncontext
import engine.profile
from engine.IO import skeletongroupparams
from engine.IO import meshmenu
from engine.IO import meshparameters
import engine.mesh
import types
from common import debug
from engine.bdycondition import BC

bcmenu = oofmenu.OOFMenuItem('Boundary_Conditions', cli_only=1)

meshmenu.meshmenu.addItem(bcmenu)


# Builder is very simple, it just calls the registration.  This
# function is the callback for all the BC class registration entries.
# Parameters can be discarded because they are shared.
def _bcbuild(menuitem, name, mesh, **params):
    bc = menuitem.data()
    bc.add_to_mesh(name, mesh)
    switchboard.notify("mesh changed", engine.mesh.meshes[mesh])
    return bc

# Boundary condition menu operations. From the point of view of the
# user interface, boundary conditions are created and assigned to a
# boundary (along with a profile) all in a single operation.  They can
# also (later) be edited in place, copied out, and so forth.
newbcmenu = bcmenu.addItem(oofmenu.OOFMenuItem('New'))

def makeMenu(menu):
    for reg in bdycondition.BC.registry:
        try:
            help = reg.tip
        except AttributeError:
            help = None
        menuitem = menu.addItem(
            oofmenu.OOFMenuItem(
            utils.space2underscore(reg.name()),
            callback=_bcbuild,
            help=help,
            params=bdycondition.bccontextgroup + reg.params) )
        menuitem.data = reg

switchboard.requestCallback(bdycondition.BC, makeMenu, newbcmenu)

makeMenu(newbcmenu)

def _bcrename(self, mesh, bc, name):
    meshcontext = engine.mesh.meshes[mesh]
    meshcontext.renameBdyCondition(bc, name)

bcmenu.addItem(oofmenu.OOFMenuItem(
    "Rename",
    callback=_bcrename,
    params=parameter.ParameterGroup(
    whoville.WhoParameter("mesh", engine.mesh.meshes),
    parameter.StringParameter('bc', tip='Old name of the boundary condition'),
    parameter.StringParameter('name', tip='New name of the boundary condition')
##     parameter.AutomaticNameParameter('name', value=automatic.automatic,
##                                      tip="New name for the bc",
##                                      resolver=bdycondition.bcnameResolver)
    ),
    help="Give a boundary condition a new name."
    ))

# Remove a boundary condition from the given mesh.
def _bcremove(self, mesh, name):
    meshcontext = engine.mesh.meshes[mesh]
    meshcontext.rmBdyCondition(name)
    switchboard.notify("mesh changed", engine.mesh.meshes[mesh])
    
bcmenu.addItem(oofmenu.OOFMenuItem(
    "Delete",
    callback=_bcremove,
    params = [bdycondition.meshparam,
              parameter.StringParameter('name')],
    help="Remove this condition from the mesh."
    ))


# Copy a boundary condition to a new boundary (for possibly another mesh).
def _bccopy(self, current, mesh, bc, name, boundary):
    meshcontext = engine.mesh.meshes[current]
    bndycond = meshcontext.getBdyCondition(bc)
    newbc = bndycond.copy(boundary)
    newbc.add_to_mesh(name, mesh)
    switchboard.notify("mesh changed", engine.mesh.meshes[mesh])
    
bcmenu.addItem(oofmenu.OOFMenuItem(
    "Copy",
    callback=_bccopy,
    params=parameter.ParameterGroup(
    whoville.WhoParameter("current", engine.mesh.meshes,
                          tip="Copy the boundary condition from this mesh."),
    whoville.WhoParameter("mesh", engine.mesh.meshes,
                          tip="Copy the boundary condition to this mesh."),
    parameter.StringParameter("bc", tip="Boundary condition being copied"),
    parameter.AutomaticNameParameter("name", value=automatic.automatic,
                                     resolver=bdycondition.bcnameResolver),
    skeletongroupparams.BoundaryParameter(
    "boundary",tip="Boundary to which the copied BC will apply."))))

# Copy all the boundary conditions from one mesh to another, copying
# them to the same boundaries, of course.
def _bccopyall(self, current, mesh):
    meshcontext = engine.mesh.meshes[current]
    allbcs = meshcontext.allBoundaryConds()
    for (name, cond) in allbcs:
        newbc = cond.copy(cond.boundary)
        newbc.add_to_mesh(name, mesh)
    switchboard.notify("mesh changed", engine.mesh.meshes[mesh])
    
bcmenu.addItem(oofmenu.OOFMenuItem(
    "Copy_All",
    callback=_bccopyall,
    params=[
    whoville.WhoParameter("current", engine.mesh.meshes,
                          tip="Copy all the boundary conditions to a mesh."),
    whoville.WhoParameter("mesh", engine.mesh.meshes,
                          tip="Target mesh for the boundary conditions.")],
    help="Copy all the boundary conditions from the current mesh to another."))


def _bcedit(menuitem, name, mesh, condition):
    meshctxt = engine.mesh.meshes[mesh]
    meshctxt.rmBdyCondition(name)
    condition.add_to_mesh(name, mesh)
    switchboard.notify("mesh changed", engine.mesh.meshes[mesh])    
    
bcmenu.addItem(oofmenu.OOFMenuItem(
    "Edit",
    callback=_bcedit,
    params = [parameter.StringParameter("name"), # No resolver, name must exist
    whoville.WhoParameter("mesh", engine.mesh.meshes),
    parameter.RegisteredParameter("condition", bdycondition.BC)],
    help="Change the attributes of a boundary condition."))

                 
    


##############################
### Similar for Profiles.
##############################

pfmenu = oofmenu.OOFMenuItem('Profiles', cli_only=1)

meshmenu.meshmenu.addItem(pfmenu)


# Profiles are special in that they're allowed to not have names in
# general, but not when called through the menu -- consequently, the
# name resolver should allow "None", and the name parameter should be
# able to provide it, but the menu callback should prohibit it.
def _pfbuild(menuitem, **kwargs):
    prof = menuitem.data() # Params already set, of course.
    name = kwargs['name']
    if name:
        prof.set_name(name)


def pfnameResolver(param, name):
    if name:
        if param.automatic():
            basename = 'profile'
        else:
            basename = name
        return engine.profile.AllProfiles.uniqueName(basename)
    return None # Default behavior, not actually required.

pfnameparam = parameter.AutomaticNameParameter(
    'name', value=automatic.automatic, tip="Select a name.",
    resolver=pfnameResolver)


newpfmenu = pfmenu.addItem(oofmenu.OOFMenuItem('New'))

def pfmakeMenu(menu):
    # Use the special "CoreProfile" class so NamedProfiles aren't in the menu.
    for r in engine.profile.CoreProfile.registry:
        menuitem = menu.addItem(
            oofmenu.OOFMenuItem(utils.space2underscore(r.name()),
                                callback = _pfbuild,
                                params= [pfnameparam] + r.params) )
        menuitem.data = r

switchboard.requestCallback(engine.profile.Profile, pfmakeMenu, newpfmenu)

pfmakeMenu(newpfmenu)


# Remove a profile from the profile manager.
def _pfremove(self, name):
    del engine.profile.AllProfiles[name]
    
pfmenu.addItem(oofmenu.OOFMenuItem(
    "Delete",
    callback=_pfremove,
    params = [parameter.StringParameter('name')]
    ))


def _pfrename(self, profile, name):
    engine.profile.AllProfiles.rename(profile, name)

pfmenu.addItem(oofmenu.OOFMenuItem(
    "Rename",
    callback=_pfrename,
    params = [
    parameter.StringParameter('profile'),
    parameter.StringParameter('name', tip='New name of the profile')]
##     parameter.AutomaticNameParameter('name', resolver=pfnameResolver)
    ))


def _pfcopy(self, profile, name):
    old_profile = engine.profile.AllProfiles[profile]
    new_profile = old_profile.clone()
    new_profile.set_name(name)

pfmenu.addItem(oofmenu.OOFMenuItem(
    "Copy",
    callback=_pfcopy,
    params = [
    parameter.StringParameter('profile'),
    parameter.AutomaticNameParameter('name', resolver=pfnameResolver) ] ) )


def _pfedit(self, name, profile):
    engine.profile.AllProfiles.replaceProfile(name, profile)
    
pfmenu.addItem(oofmenu.OOFMenuItem(
    "Edit",
    callback=_pfedit,
    params = [parameter.StringParameter("name"), # No resolver, name must exist
    parameter.RegisteredParameter("profile", engine.profile.CoreProfile)]))

