# -*- python -*-
# $RCSfile: skeletonIO.py,v $
# $Revision: 1.21 $
# $Author: haan $
# $Date: 2004/09/09 20:54:42 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# Menu commands for reading a Skeleton from a file.

# TODO LATER: the name of the Skeleton appears many times in the file,
# in each menu command.  That makes it hard for the user to edit the
# file to change the name.  Ugly solution: have a global
# "currentskeleton" variable which is set by the initial Skeleton.New
# command.  Difficult solution: allow variables to be set locally *in*
# the data file, and store the skeleton name there.

from common.IO import mainmenu
from common.IO import oofmenu
from common.IO import parameter
from common.IO import whoville
from engine import skeletoncontext
from engine import skeletonboundary
import common.microstructure
import engine.skeleton
from common import debug
from SWIG.common import switchboard

OOFMenuItem = oofmenu.OOFMenuItem

OOF = mainmenu.OOF

skelmenu = OOF.LoadData.addItem(OOFMenuItem('Skeleton'))

def _newSkel(menuitem, name, microstructure):
    engine.skeleton.newEmptySkeleton(name, microstructure)
    
skelmenu.addItem(OOFMenuItem(
    'New',
    callback=_newSkel,
    params=[parameter.StringParameter('name'),
            whoville.WhoParameter('microstructure',
                         common.microstructure.microStructures)]))

###

def _loadNodes(menuitem, skeleton, points):
    # read nodes as (x,y) tuples of floats
    skelcontext = skeletoncontext.skeletonContexts[skeleton]
    skeleton = skelcontext.getObject()
    for node in points:
        skeleton.newNode(*node)
    switchboard.notify(('who changed', 'Skeleton'), skelcontext)

skelmenu.addItem(OOFMenuItem(
    'Nodes',
    callback=_loadNodes,
    params=[whoville.WhoParameter('skeleton', skeletoncontext.skeletonContexts),
            parameter.ListOfTuplesOfFloatsParameter('points')]
    ))

###

def _loadElements(menuitem, skeleton, nodes):
    # read elements as tuples of node indices
    skelcontext = skeletoncontext.skeletonContexts[skeleton]
    skeleton = skelcontext.getObject()
    for nodelist in nodes:
        skeleton.loadElement(*nodelist)
    skelcontext.getTimeStamp(None).increment()
    switchboard.notify(('who changed', 'Skeleton'), skelcontext)

skelmenu.addItem(OOFMenuItem(
    'Elements',
    callback=_loadElements,
    params=[whoville.WhoParameter('skeleton', skeletoncontext.skeletonContexts),
            parameter.ListOfTuplesOfIntsParameter('nodes')]
    ))

###

def _loadPinnedNodes(menuitem, skeleton, nodes):
    # "nodes" is a list of integer indices into skeleton.nodes
    skelcontext = skeletoncontext.skeletonContexts[skeleton]
    skeleton = skelcontext.getObject()
    pinned = skelcontext.pinnednodes
    pinned.pin([skeleton.nodes[i] for i in nodes])
    pinned.signal()
skelmenu.addItem(OOFMenuItem(
    'PinnedNodes',
    callback = _loadPinnedNodes,
    params = [
    whoville.WhoParameter('skeleton', skeletoncontext.skeletonContexts),
    parameter.ListOfIntsParameter('nodes')]
    ))

###
    
def _loadPointBoundary(menuitem, skeleton, name, nodes, exterior):
    # "nodes" is a list of integer indices into skeleton.nodes
    skelcontext = skeletoncontext.skeletonContexts[skeleton]
    skeleton = skelcontext.getObject()
    skelcontext.createPointBoundary(name, [skeleton.nodes[i] for i in nodes],
                                    exterior, autoselect=0)
    # The last one, "autoselect" blocks the auto-selection feature
    # in the skeleton boundary page.

skelmenu.addItem(OOFMenuItem(
    'PointBoundary',
    callback=_loadPointBoundary,
    params=[
    whoville.WhoParameter('skeleton', skeletoncontext.skeletonContexts),
    parameter.StringParameter('name'),
    parameter.ListOfIntsParameter('nodes'),
    parameter.IntParameter('exterior', 0)]
    ))

###

def _loadEdgeBoundary(menuitem, skeleton, name, edges, exterior):
    # "edges" is a list of pairs of integer indices into skeleton.nodes
    skelcontext = skeletoncontext.skeletonContexts[skeleton]
    skeleton = skelcontext.getObject()
    seglist = [skeleton.getSegment(
        *[skeleton.nodes[i] for i in e] )
               for e in edges]
    startnode = skeleton.nodes[edges[0][0]]
    skelcontext.createEdgeBoundary(name, seglist, startnode,
                                   exterior, autoselect=0)
    # The last one, "autoselect" blocks the auto-selection feature
    # in the skeleton boundary page.

    
skelmenu.addItem(OOFMenuItem(
    'EdgeBoundary',
    callback=_loadEdgeBoundary,
    params=[
    whoville.WhoParameter('skeleton', skeletoncontext.skeletonContexts),
    parameter.StringParameter('name'),
    parameter.ListOfTuplesOfIntsParameter('edges'),
    parameter.IntParameter('exterior', 0)]
    ))

#############

def writeSkeleton(datafile, skeletoncontext):
    skeletoncontext.begin_reading()
    try:
        skeleton = skeletoncontext.getObject()
        skelpath = skeletoncontext.path()

        # Create skeleton.
        datafile.startCmd(skelmenu.New)
        datafile.argument('name', skeletoncontext.name())
        datafile.argument('microstructure', skeleton.MS.name())
        datafile.endCmd()

        # Define nodes.
        datafile.startCmd(skelmenu.Nodes)
        datafile.argument('skeleton', skelpath)
        datafile.argument('points', [(nd.position().x, nd.position().y)
                                     for nd in skeleton.nodes])
        datafile.endCmd()

        # Nodes are written and read in the order in which they're stored
        # in the Skeleton's nodes list.  This is not the same as the
        # nodes' internal index, so we need to create a dictionary of
        # indices in the nodes list so that the elements' node tuples are
        # correct.
        nodedict = {}
        c = 0
        for node in skeleton.nodes:
            nodedict[node] = c
            c += 1

        # Define elements, using the node indices in nodedict.
        datafile.startCmd(skelmenu.Elements)
        datafile.argument('skeleton', skelpath)
        datafile.argument('nodes', [tuple([nodedict[node] for node in el.nodes])
                                    for el in skeleton.elements])
        datafile.endCmd()

        # Pinned nodes
        datafile.startCmd(skelmenu.PinnedNodes)
        datafile.argument('skeleton', skelpath)
        datafile.argument('nodes', [nodedict[node] for node in skeletoncontext.pinnednodes.retrieve()])
        datafile.endCmd()

        # Point boundaries
        for pbname, pbdy in skeleton.pointboundaries.items():
            datafile.startCmd(skelmenu.PointBoundary)
            datafile.argument('skeleton', skelpath)
            datafile.argument('name', pbname)
            datafile.argument('nodes', [nodedict[node] for node in pbdy.nodes])
            exterior = 0
            if isinstance(pbdy, skeletonboundary.ExteriorSkeletonPointBoundary):
                exterior = 1
            datafile.argument('exterior', exterior)
            datafile.endCmd()

        # Edge boundaries
        for ebname, ebdy in skeleton.edgeboundaries.items():
            datafile.startCmd(skelmenu.EdgeBoundary)
            datafile.argument('skeleton', skelpath)
            datafile.argument('name', ebname)
            edgeset = rearrangeEdges([
                tuple([nodedict[x] for x in edge.get_nodes()])
                for edge in ebdy.edges
                ])
            datafile.argument('edges', edgeset)
            exterior = 0
            if isinstance(ebdy, skeletonboundary.ExteriorSkeletonEdgeBoundary):
                exterior = 1
            datafile.argument('exterior', exterior)
            datafile.endCmd()
    finally:
        skeletoncontext.end_reading()
        
###########################################################

# A function to put edges in a right order.
def rearrangeEdges(edges):
    # 'edges' is a list of tuples of integers.  Each tuple is an edge,
    # and each integer is a node index (from nodedict in
    # writeSkeleton, above).
    if len(edges) == 1:
        return edges

    # Construct dictionaries of edges keyed by their starting and end points.
    tails = {}           
    heads = {}         
    for edge in edges:
        heads[edge[0]] = edge
        tails[edge[1]] = edge

    newedges = [edges[0]]

    # Build the list of contiguous edges going backwards from the
    # start point of edges[0].
    lastpt = edges[0][0]
    try:
        while 1:
            next = tails[lastpt]
            newedges.append(next)
            lastpt = next[0]
    except KeyError:
        pass

    newedges.reverse()

    # Extend the list of contiguous edges going forwards from the end
    # point of edges[0].
    lastpt = edges[0][1]
    try:
        while 1:
            next = heads[lastpt]
            newedges.append(next)
            lastpt = next[1]
    except KeyError:
        pass

    return newedges
        
        

    
