# -*- python -*-
# $RCSfile: skeletonmenu.py,v $
# $Revision: 1.86 $
# $Author: redwing $
# $Date: 2004/09/17 21:16:24 $


# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 


# Mesh menu.  Contains the commands for creating and refining/modifying
# skeletons, and transforming them to real meshes.


from SWIG.common import switchboard
from common import debug
from common import enum
from common import labeltree
from common import microstructure
from common import quit
from common.IO import automatic
from common.IO import datafile
from common.IO import mainmenu
from common.IO import microstructureIO
from common.IO import oofmenu
from common.IO import parameter
from common.IO import reporter
from common.IO import whoville
from engine import skeleton
from engine import skeletoncontext
from engine import skeletonmodifier
from engine.IO import skeletonIO


OOF = mainmenu.OOF
Parameter = parameter.Parameter
IntParameter = parameter.IntParameter
WhoParameter = whoville.WhoParameter
StringParameter = parameter.StringParameter
microStructures = microstructure.microStructures
SkeletonModifier = skeletonmodifier.SkeletonModifier
AutomaticNameParameter = parameter.AutomaticNameParameter

skeletonmenu = OOF.addItem(oofmenu.OOFMenuItem('Skeleton', cli_only=1))

#####################

def skeletonNameResolver(param, startname):
    if param.automatic():
        basename = 'skeleton'
    else:
        basename = startname
    msname = param.group['microstructure'].value
    mspath = labeltree.makePath(msname)
    return skeletoncontext.skeletonContexts.uniqueName(mspath + [basename])

######################

def _skeleton_from_mstructure(menuitem, name, microstructure, x_elements,
                               y_elements, skeleton_geometry):
    pBar = menuitem.getProgressBar("continuous")
    reporter.start_progressbar()
    try:
        ms = microStructures[microstructure].getObject()
        skel = skeleton.initialSkeleton(name, x_elements, y_elements,
                                        ms, skeleton_geometry)
    finally:
        reporter.progressbar_completed()
        reporter.end_progressbar()

skeletonmenu.addItem(
    oofmenu.OOFMenuItem('New',
                        callback=_skeleton_from_mstructure,
                        threadable=oofmenu.THREADABLE,
                        params=parameter.ParameterGroup(
    AutomaticNameParameter('name',
                           value=automatic.automatic,
                           resolver=skeletonNameResolver,
                           tip="Name of the new skeleton."),
    WhoParameter('microstructure', microStructures),
    IntParameter('x_elements', 4),
    IntParameter('y_elements', 4),
    Parameter('skeleton_geometry', skeleton.SkeletonGeometry, skeleton.QuadSkeleton())
    ),
                        help="Create a new Skeleton in a Microstructure."
                         ))

#########################

def _simple_skeleton_from_ms(menuitem, name, microstructure, skeleton_geometry):
    ms = microStructures[microstructure].getObject()
    x_elements = ms.sizeInPixels()[0]
    y_elements = ms.sizeInPixels()[1]
    skel = skeleton.initialSkeleton(name, x_elements, y_elements,
                                    ms, skeleton_geometry)

skeletonmenu.addItem(
    oofmenu.OOFMenuItem('Simple',
                        callback=_simple_skeleton_from_ms,
                        params=parameter.ParameterGroup(
    AutomaticNameParameter('name',
                           value=automatic.automatic,
                           resolver=skeletonNameResolver,
                           tip="Name of the simple skeleton."),
    WhoParameter('microstructure', microStructures),
    Parameter('skeleton_geometry', skeleton.SkeletonGeometry, skeleton.QuadSkeleton())
    ),
                        help="Create a new Skeleton with one quadrilateral or two triangular elements per pixel of the Microstructure."
                        ))

#########################

def _skeleton_delete(menuitem, skeleton):
    skeletoncontext.skeletonContexts.remove(skeleton)


skeletonmenu.addItem(
    oofmenu.OOFMenuItem('Delete',
                        callback=_skeleton_delete,
                        params=[
    WhoParameter('skeleton', skeletoncontext.skeletonContexts)],
                        help="Delete a Skeleton from a Microstructure."))


#########################

def _skeleton_copy(menuitem, skeleton, name):
    # skeleton is a colon separated string
    oldskelpath = labeltree.makePath(skeleton)
    oldskelcontext = skeletoncontext.skeletonContexts[skeleton]
    if name is automatic.automatic:
        nm = skeletoncontext.skeletonContexts.uniqueName(skeleton)
    else:
        nm = name
        
    orig = oldskelcontext.getObject()
    newskel = orig.properCopy()
    for e in oldskelcontext.edgeboundaries.values():
        newskel.mapBoundary(e, orig)
    for p in oldskelcontext.pointboundaries.values():
        newskel.mapBoundary(p, orig)

    msname = oldskelpath[0]
    # "add" calls the SkeletonContext constructor, which calls
    # "disconnect" on the skeleton, ensuring that the propagation
    # we just did doesn't mess up the old skeleton context.
    skeletoncontext.skeletonContexts.add([msname, nm], newskel)

    newskelcontext = skeletoncontext.skeletonContexts[ [msname, nm] ]
    newskelcontext.groupCopy(oldskelcontext)


def copySkeletonNameResolver(param, startname):
    if param.automatic():
        basename = 'skeleton'
    else:
        basename = startname
    skelname = param.group['skeleton'].value
    skelcontext = skeletoncontext.skeletonContexts[skelname]
    return skelcontext.uniqueName(basename)

skeletonmenu.addItem(oofmenu.OOFMenuItem('Copy',
                                         callback=_skeleton_copy,
                                         params=parameter.ParameterGroup(
    WhoParameter('skeleton', skeletoncontext.skeletonContexts),
    AutomaticNameParameter('name',
                           resolver=copySkeletonNameResolver,
                           value=automatic.automatic,
                           tip="Name of the copy")
    ),
                                         help="Copy a Skeleton within a Microstructure."
                                         ))


########################

def _skeleton_rename(menuitem, skeleton, name):
    # skeleton is a colon separated string
    oldskelpath = labeltree.makePath(skeleton)
    skel = skeletoncontext.skeletonContexts[oldskelpath]
    skel.rename(name, exclude=oldskelpath[-1])

skeletonmenu.addItem(oofmenu.OOFMenuItem('Rename',
                                         callback=_skeleton_rename,
                                         params=parameter.ParameterGroup(
    WhoParameter('skeleton', skeletoncontext.skeletonContexts),
    parameter.StringParameter('name', '',
                              tip='New name for the skeleton')
    ),
                                         help="Rename a Skeleton."
                                         ))


########################

def _modify(menuitem, skeleton, modifier):
    ModifyProgressBar = menuitem.getProgressBar(modifier.get_progressbar_type())
    context = skeletoncontext.skeletonContexts[skeleton]
    ## context is actually the Who of the skeleton
    context.reserve()
    try:
        context.begin_writing()
        try:
            reporter.start_progressbar()
            skel = modifier.apply(context.getObject(), context)
            ## parallelize for each modifier
            ## Each modifier should put 
            ## a skeleton in the POM before returning.
            ## This skeleton will be 
            ## retrieved by the back-end and then
            ## pushed into the stack
            reporter.progressbar_completed()
            reporter.end_progressbar()
            ## skel is None whenever the modifier fails
            ## or is interrupted from finishing its task
            if skel is None:
                reporter.warn("Modify Process Interrupted")
                return
                ## return will force all the finally's to be executed
            context.pushModification(skel) ## parallelized
            skel.needsHash()
        finally:
            context.end_writing()
        reporter.start_progressbar()
        modifier.postProcess(context) ## parallelize for each modifier
        reporter.progressbar_completed()
        reporter.end_progressbar()
        ## If the skeleton is modified in postProcess, use
        ## begin/end_writing inside the function call to guarantee
        ## that no dead-locking occurs because of possible switchboard
        ## calls to READ or REDRAW that may make use of
        ## begin/end_reading(). See anneal.py for an example.
    finally:
        ## guarantee that the reservation is cancelled even if an
        ## exception is raised
        context.cancel_reservation()

        if ModifyProgressBar.query_stop() or ModifyProgressBar.get_success()<0:
            ModifyProgressBar.set_failure()
            ModifyProgressBar.set_message("Failed")
            return None
        else:
            ModifyProgressBar.set_success()
            ModifyProgressBar.set_message("Succeeded")
        switchboard.notify('redraw')
        switchboard.notify('Skeleton modified', skeleton, modifier)



skeletonmenu.addItem(
    oofmenu.OOFMenuItem("Modify",
                        callback=_modify,
                        threadable=oofmenu.THREADABLE,
                        params=[WhoParameter('skeleton',
                                             skeletoncontext.skeletonContexts),
                                Parameter('modifier',SkeletonModifier,None)],
                        help="Modify a Skeleton in various ways."
                        ))

########################

def _undo(menuitem, skeleton):
    context = skeletoncontext.skeletonContexts[skeleton]
    if context.undoable():
        context.begin_writing()
        try:
            context.undoModification()
        finally:
            context.end_writing()
            switchboard.notify('redraw')
    else:
        reporter.report("Can't undo skeleton-modification.")

skeletonmenu.addItem(
    oofmenu.OOFMenuItem("Undo",
                        callback=_undo,
                        params=[WhoParameter('skeleton',
                                skeletoncontext.skeletonContexts)],
                        help="Undo a Skeleton modification."))
        
def _redo(menuitem, skeleton):
    context = skeletoncontext.skeletonContexts[skeleton]
    if context.redoable():
        context.begin_writing()
        try:
            context.redoModification()
        finally:
            context.end_writing()
            switchboard.notify('redraw')
    else:
        reporter.report("Can't redo skeleton-modification.")

skeletonmenu.addItem(
    oofmenu.OOFMenuItem("Redo",
                        callback=_redo,
                        params=[WhoParameter('skeleton',
                                skeletoncontext.skeletonContexts)],
                        help="Redo a Skeleton modification."))

#############################

def sanity_check(menuitem, skeleton, quick):
    skel = skeletoncontext.skeletonContexts[skeleton].getObject()
    if quick:
        return skel.quick_sanity_check()
    else:
        return skel.sanity_check()

OOF.Help.Debug.addItem(oofmenu.OOFMenuItem(
    'Sanity_Check',
    callback=sanity_check,
    ordering=100,
    params=[WhoParameter('skeleton', skeletoncontext.skeletonContexts),
            parameter.BooleanParameter('quick', 1)],
    help="Check for errors in Skeleton construction."))

#####################

def saveSkeleton(menuitem, filename, mode, format, skeleton):
    quit.disallow()
    try:
        skelcontext = skeletoncontext.skeletonContexts[skeleton]
        dfile = datafile.writeDataFile(filename, mode, format)
        microstructureIO.writeMicrostructure(
            dfile,
            microstructure.getMSContextFromMS(skelcontext.getObject().MS))
        skeletonIO.writeSkeleton(dfile, skelcontext)
        dfile.close()
    finally:
        quit.allow()

OOF.File.Save.addItem(oofmenu.OOFMenuItem(
    'Skeleton',
    callback = saveSkeleton,
    params = [
    parameter.StringParameter('filename'),
    parameter.StringParameter('mode'),
    enum.EnumParameter('format', datafile.DataFileFormat, datafile.ASCII),
    whoville.WhoParameter('skeleton', skeletoncontext.skeletonContexts)],
    help="Save a Skeleton to a file."
    ))

