// -*- C++ -*-
// $RCSfile: femesh.h,v $
// $Revision: 1.38 $
// $Author: langer $
// $Date: 2004/08/12 17:52:12 $


/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#ifndef FEMESH_H
#define FEMESH_H

class ElementIterator;
class FuncNodeIterator;
class FEMesh;
class NodeIterator;
class Solver;

#include <Python.h>

#include "common/lock.h"
#include "common/timestamp.h"
#include "engine/matvec.h"
#include "common/pythonexportable.h"
#include <string>
#include <vector>

class CMicrostructure;
class CompoundField;
class Coord;
class CSRmatrix;
class DegreeOfFreedom;
class Element;
class Equation;
class Field;
class Flux;
class NodalEquation;
class Node;
class FuncNode;
class StiffnessMatrix;


//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// MeshIterators

enum NodeIteratorType {NI_ALL, NI_SELECTED, NI_UNSELECTED};
enum ElementIteratorType {EI_ALL, EI_SELECTED, EI_UNSELECTED};

class NodeIterator {
private:
  const FEMesh * const mesh;
  NodeIteratorType type;
  int index;
public:
  NodeIterator(const FEMesh * const, NodeIteratorType type=NI_ALL);
  void operator++();
  bool begin() const;
  bool end() const;
  Node *node() const;
  friend class FEMesh;
};

class FuncNodeIterator {
private:
  const FEMesh * const mesh;
  NodeIteratorType type;
  int index;
public:
  FuncNodeIterator(const FEMesh * const, NodeIteratorType type=NI_ALL);
  void operator++();
  bool end() const;
  FuncNode *node() const;
  friend class FEMesh;
};

class ElementIterator {
private:
  const FEMesh * const mesh;
  int index;
  ElementIteratorType type;
public:
  ElementIterator(const FEMesh* const, ElementIteratorType type=EI_ALL);
  void operator++();
  bool end() const;
  int size() const;
  Element *element() const;
  friend class FEMesh;
};

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//




class FEMesh : public PythonExportable<FEMesh> {
  // FEMesh has to be derived from PythonExportable so that it can be
  // used as an argument in switchboard messages.
private:
  CMicrostructure * microstructure;
  RWLock * rwlock;
public:
  FEMesh();
  FEMesh(CMicrostructure *); //for super annealing only
  virtual ~FEMesh();
  CMicrostructure * get_microstructure() const {return microstructure;}
  virtual const std::string &classname() const;
  virtual const std::string &modulename() const;

  Node *operator[](const NodeIterator &) const;
  FuncNode *operator[](const FuncNodeIterator &) const;
  Element *operator[](const ElementIterator&) const;

  ElementIterator element_iterator(ElementIteratorType = EI_ALL) const;
  NodeIterator node_iterator(NodeIteratorType = NI_ALL) const;
  FuncNodeIterator funcnode_iterator(NodeIteratorType = NI_ALL) const;
  void renumber_nodes();
  void renumber_elements();

  Node *newMapNode(const Coord&); // the only way to make a Node
  FuncNode *newFuncNode(const Coord&); // the only way to make a FuncNode
  void AddElement(Element*);

  // Caution: NodeIterator::index is not necessarily the same as
  // node.index().  They are only the same if FEMesh::renumber_nodes()
  // has been called.  The argument to FEMesh::getNode() is the
  // NodeIterator::index.
  Node *getNode(int) const;

  // Temporary function for finding the closest node.
  Node *closestNode(const double x, const double y);

  DegreeOfFreedom *CreateDoF(double x=0); // only way to make a DoF
  void RemoveDoF(DegreeOfFreedom*);
  void activate_active_dofs(FuncNode *) const;

  void dump_dof() const;	// for debugging
  void dump_eqn() const;


  // NodalEquations represent one component of one Equation at a Node.
  // They are to the rows of the stiffness matrix what the
  // DegreeOfFreedoms are to the columns.
  NodalEquation *CreateNodalEqn();
  void RemoveNodalEqn(NodalEquation*);

//   void precompute();
  void requirePrecompute();	// call when Materials change or Fields defined
  Lock precomputeLock;
  bool precomputeRequired;

  void make_stiffness();
  const StiffnessMatrix &stiffness_matrix() const;
  VECTOR_D &boundary_rhs() const;
  void boundary_rhs_init(int);
  void dump_boundary_rhs();
  bool solve(Solver *);
//   bool test(int *i, double *x, int k) const {
//     *i = *i + 1;
//     *x = *x * k;
//     return 1;
//   }

  // Run post-solver processing routines in those materials/properties
  // that have them.  This should eventually depend which auxiliary
  // fields are contributed to by the material.
  void post_process();

  std::vector<int> freedofmap;
  std::vector<int> fixeddofmap;
  std::vector<int> indepeqnmap;
  std::vector<int> depeqnmap;
  void rcmapper(bool (FEMesh::*)(int) const, std::vector<int>&, int) const; 
  void mapdofeqs();  // Creates the matrix-partitioning maps.
  void resetdofeqs(); // Undoes BC invocation, sets dofs/eqns to free/indep.
  void cleanlists();
  CSRmatrix Amatrix() const;
  CSRmatrix Bmatrix() const;
  CSRmatrix Cmatrix() const;
  CSRmatrix Cpmatrix() const;
  
  Element *getElement(int i) const; //{ return element[i]; }

  void refreshMaterials(PyObject *skeleton);

  int nnodes() const;
  int nelements() const;

  void activate_equation(Equation &);
  void deactivate_equation(Equation &);
  bool is_active_equation(const Equation&) const;
  int active_equation_index(const Equation&) const;
  int n_active_eqns() const; // { return n_active_eqn; }

  void activate_flux(const Flux&);
  void deactivate_flux(const Flux&);
  bool is_active_flux(const Flux&) const;

  // activate, etc, for Fields have to be routed through Field virtual
  // functions, because they have to know if the Field is a
  // CompoundField or not.  The routines that do the real work are
  // FEMesh::do_activate, etc.
  void activate_field(Field&);
  void deactivate_field(Field&);
  void acquire_field_data(Field &field, const FEMesh*);
  Field &define_field(Field&);
  void undefine_field(Field&);
  void set_in_plane(const Field&, bool);
  bool in_plane(const Field&) const;

  // active fields are being solved for
  void do_activate_field(Field&);
  void do_deactivate_field(Field&);
  bool is_active_field(const Field&) const;
  //  int n_active_dofs() const { return n_active_dof; } // # active dofs per node
  // defined fields have values at nodes
  void do_define_field(Field&);
  void do_undefine_field(Field&);
  bool is_defined_field(const Field&) const;
  int defined_index(const Field&) const; // offset in node's dof list
  int n_defined_dofs() const;// { return n_defined_dof; } // # of dofs per node

  // List of this mesh's defined fields.
  std::vector<CompoundField*>* all_fields() const;
  std::vector<Equation*>* all_equations() const;
  std::vector<Flux*>* all_fluxes() const;

  // API for grabbing a reference to the RWLock
  inline RWLock  * get_rwlock() { return rwlock;};

private:
  // These lists can be accessed through the MeshIterators.
  std::vector<FuncNode*> funcnode;// nodes at which dofs are defined
  std::vector<Node*> mapnode;	// nodes only used for geometry mapping
  std::vector<Element*> element;
  int ncount;

  // Master lists of degrees of freedom and nodal equations
  std::vector<DegreeOfFreedom*> dof;
  std::vector<NodalEquation*> nodaleqn;
  void housekeeping();		     // do garbage collection on the lists
  bool dof_list_needs_cleaning;
  bool nodaleqn_list_needs_cleaning; // do the lists need to be cleaned up?
  void cleanmap(std::vector<int>&);       // Clean up a particular map.
  // Above routine is called by public method "cleanlists".

  // Is a Flux active?  Stored as an int, because a Flux can be
  // activated more than once.
  std::vector<int> active_flux;

  // This is data that can't live in the Field class, because there is
  // only one instance of each Field, but it might have different
  // properties on each FEMesh, and there might be more than one FEMesh.
  class FieldData {
  public:
    FieldData();
  private:
    //    int active_index;		// used to order the active dofs
    int defined_index;		// used to compute location in a node's doflist
    bool active;		// is this field being solved for?
    bool defined;		// have DoF's been assigned?
    bool in_plane;		// is this field in plane?
    friend class FEMesh;
  };
  std::vector<FieldData> fielddata;

  // This is data that can't live in the Equation class, for the same
  // reason that the FieldData can't live in the Field class.
  class EquationData {
  public:
    EquationData();
  private:
    int active_index;
    bool active;
    friend class FEMesh;
  };
  std::vector<EquationData> eqndata;
  void clean_nodaleqn();
  void clean_doflist();

  int n_defined_dof;		// number of dofs defined at each node
  //  int n_active_dof;		// number of active dofs at each node
  int n_active_eqn;		// number of active eqns at each node


  StiffnessMatrix *stiffness;
  VECTOR_D *boundary_right_side;  // Boundary's contributions to forces.
  VECTOR_D *body_force_rhs;       // RHS contribs from integrating properties

  // Routines used to pick rows and columns of the stiffness matrix
  // when extracting submatrices
  bool solve_for_this_dof(int) const;
  bool solve_this_nodal_eqn(int) const;
  bool dont_solve_for_this_dof(int) const;
  bool dont_solve_this_nodal_eqn(int) const;

  // Routines for extracting parts of the rhs and solution vectors
public:
  StiffnessMatrix * get_stiffness() const {return stiffness;}
  void get_unknowns(VECTOR_D&, const std::vector<int>&) const;
  void get_rhs(VECTOR_D&, const std::vector<int>&) const;
  // The opposites of the above routines...
  void set_unknowns(const VECTOR_D&, const std::vector<int>&);
  void set_rhs(const VECTOR_D&, const std::vector<int>&);
private:
  friend class Node;
  friend class NodeIterator;
  friend class FuncNodeIterator;
  friend class ElementIterator;
};


#endif // FEMESH_H

