// -*- C++ -*-
// $RCSfile: pixelintersection.C,v $
// $Revision: 1.5 $
// $Author: langer $
// $Date: 2003/04/11 15:57:14 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#include "common/geometry.h"
#include "engine/cskeleton.h"

#include <math.h>
#include <vector>
#include <algorithm>

static inline double triangleArea(const Coord &v0, const Coord &v1,
				  const Coord &v2)
{
  return 0.5*(v1 - v0)%(v2 - v0);
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// The ConvexPolygon class is used in pixelIntersection() to hold the
// intersection of a pixel and a polygon.  It differs from CPolygon in
// that it figures out the proper order for its corners.

// Comparison object for std::sort for ConvexPolygon.  operator()
// returns true if its first argument comes before the second when
// traversing the polygon nodes counterclockwise from referencept.

class ConvexPolygonPointSorter {
private:
  const Coord referencept;
public:
  ConvexPolygonPointSorter(const Coord &refpt)
    : referencept(refpt) {}
  bool operator()(const Coord &p1, const Coord &p2) const {
    double a = triangleArea(referencept, p1, p2);
    if(a > 0)
      return true;
    return false;
  }
};

// Class for constructing a convex polygon from a set of corners.

class ConvexPolygon {
private:
  std::vector<Coord> vertex;
  bool sorted;
  int sort();
public:
  ConvexPolygon() : sorted(0) {}
//   const Coord &operator()(int i) const { return vertex[i]; }
  int npts() const { return vertex.size(); }
  void add(const Coord &pt);
  double area();
};

void ConvexPolygon::add(const Coord &pt) {
  vertex.push_back(pt);
  sorted = 0;
}

double ConvexPolygon::area() {
  if(vertex.size() < 3) return 0.0;
  if(!sort())
    return 0.0;
  double a = 0.0;
  for(int j=2; j<vertex.size(); j++)
    a += triangleArea(vertex[0], vertex[j-1], vertex[j]);
  return a;
}

int ConvexPolygon::sort() {
  // Sort vertices so that they proceed counterclockwise. The sorting
  // starts with vertex[1], not vertex[0], since the starting vertex
  // (referencept) has to be chosen arbitrarily.  If the same point
  // appears twice in the list, then the sorting algorithm can fail,
  // since it relies on the cross product of vectors drawn between
  // points. In particular, it WILL fail if the referencept is a
  // redundant point.  So first look for a nonredundant reference
  // point, then sort the vertices, then remove the redundant points.
  // This is more efficient than searching for all redundant points in
  // an unsorted list.
  if(sorted) return 1;
  if(vertex.size() < 3) {
    sorted = true;
    return 0;
  }
  // Most of the time all the vertices will be unique, so do a quick
  // check on vertex[0] first.
  int duplicate = 0;
  for(int j=1; j<vertex.size() && !duplicate; j++)
    if(vertex[0] == vertex[j])
      duplicate = j;
  
  if(!duplicate) {
    // vertex[0] is unique.  Putting vertices in order will work, even
    // if other vertices are not unique.
    std::sort(vertex.begin()+1, vertex.end(),
	      ConvexPolygonPointSorter(vertex[0]));
    // Remove duplicates
    std::vector<Coord>::iterator newend =
      std::unique(vertex.begin(), vertex.end());
    vertex.erase(newend, vertex.end());
  }
  else {
    // vertex[0] isn't unique.  We have to sort to find the redundant
    // vertices, remove them, and then put the remaining vertices in
    // counterclockwise order.

    // sort vertices alphabetically and remove redundant ones.
    std::sort(vertex.begin(), vertex.end());
    std::vector<Coord>::iterator newend = 
      std::unique(vertex.begin(), vertex.end());
    vertex.erase(newend, vertex.end());
    // put vertices in counterclockwise order.
    std::sort(vertex.begin()+1, vertex.end(),
	      ConvexPolygonPointSorter(vertex[0]));
  }
  sorted = 1;
  return 1;			// success
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Find the area of intersection of a polygon and a pixel.

// find intersection of a pixel side (p1, p2) with an polygon side
// (n1, n2) assuming that the pixel side is vertical and that the
// polygon side is not. ***The segments are closed (contain their
// endpoints).***

static int yintercept(const Coord &n1, const Coord &n2,
		      const Coord &p1, const Coord &p2,
		      Coord &intercept)

{
  // rename polygon corners so that polya is to the left of polyb
  const Coord &polya = (n1(0) < n2(0) ? n1 : n2);
  const Coord &polyb = (&polya == &n1 ? n2 : n1);
  // check that polygon side straddles the (extended) pixel side
  if(polya(0) > p1(0) || polyb(0) < p1(0))
    return 0;
  // rename pixel corners so that pa is below pb
  const Coord &pa = (p1(1) < p2(1) ? p1 : p2);
  const Coord &pb = (&pa == &p1 ? p2 : p1);
  intercept(0) = pa(0);
  intercept(1) = polya(1) +
    (polyb(1) - polya(1))*(pa(0) - polya(0))/(polyb(0) - polya(0));
  if(intercept(1) > pb(1) || intercept(1) < pa(1))
    return 0;
  return 1;
  
}

// find intersection of a pixel side (p1, p2) with a polygon side
// (n1, n2) assuming that the pixel side is horizontal and that the
// polygon side is not. ***The segments are closed (contain their
// endpoints).***

static int xintercept(const Coord &n1, const Coord &n2,
		      const Coord &p1, const Coord &p2,
		      Coord &intercept)
{
  // rename polygon corners so that polya is below polyb
  const Coord &polya = (n1(1) < n2(1) ? n1 : n2);
  const Coord &polyb = (&polya == &n1 ? n2 : n1);
  // check that polygon side straddles the (extended) pixel side
  if(polya(1) > p1(1) || polyb(1) < p1(1))
    return 0;
  // rename pixel corners so that pa is to the left of pb
  const Coord &pa = (p1(0) < p2(0) ? p1 : p2);
  const Coord &pb = (&pa == &p1 ? p2 : p1);
  intercept(1) = pa(1);
  intercept(0) = polya(0) +
    (polyb(0) - polya(0))*(pa(1) - polya(1))/(polyb(1) - polya(1));
  if(intercept(0) > pb(0) || intercept(0) < pa(0))
    return 0;
  return 1;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

double CSkeletonElement::pixel_intersection(const ICoord *pxl,
					    const Coord *pxlsize) const
{
  int ncorners = nnodes();
  ConvexPolygon intersection;

  // Representation of the pixel as a rectangle.
  double dx = (*pxlsize)(0);
  double dy = (*pxlsize)(1);
  Coord ll((*pxl)(0)*dx, (*pxl)(1)*dy);
  Coord ur(ll + Coord(dx, dy));
  CRectangle pixel(ll, ur);

  // Check corners of the pixel to see if they're in the element.
  
  for(int i=0; i<4; i++) {
    Coord c = pixel[i];
    if(interior(&c)) {
      intersection.add(c);
    }
  }
  int np = intersection.npts();	// no. of pixel corners inside element
  if(np == 4) {			// pixel is entirely within element
    return dx*dy;
  }

  // Check corners of the element to see if they're in the pixel.
  for(int i=0; i<ncorners; i++) {
    const Coord &pos = nodes[i]->position();
    if(pixel.contains(pos))
      intersection.add(pos);
  }
  int ne = intersection.npts() - np; // no. of element corners inside pixel
  if(ne == ncorners) {		// element is entirely within pixel
    return area();
  }

  // Look for intersections of edges.
  for(int i=0; i<ncorners; i++) {
    const Coord &n1 = nodes[i]->position();
    const Coord &n2 = nodes[(i+1)%ncorners]->position();
    Coord intercept;
    // Intersections with horizontal pixel edges.  It's not necessary
    // to check horizontal element edges, since if they lie along a
    // pixel edge, the intersections will have been detected when
    // checking to see if the element corners are inside the pixel.
    if(n1(1) != n2(1)) {
      if(xintercept(n1, n2, pixel[0], pixel[1], intercept))
	intersection.add(intercept);
      if(xintercept(n1, n2, pixel[3], pixel[2], intercept))
	intersection.add(intercept);
    }
    // Intersections with vertical pixel edges.
    if(n1(0) != n2(0)) {
      if(yintercept(n1, n2, pixel[0], pixel[3], intercept))
	intersection.add(intercept);
      if(yintercept(n1, n2, pixel[1], pixel[2], intercept))
	intersection.add(intercept);
    }
  }
  return intersection.area();
}

