// -*- C++ -*-
// $RCSfile: solver.C,v $
// $Revision: 1.32 $
// $Author: reida $
// $Date: 2004/09/09 18:48:31 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>
#include <math.h>
#include "femesh.h"
#include "solver.h"
#include "csrmatrix.h"
#include "common/smallmatrix.h"
#include "common/IO/progressbar.h"
#include "matvec.h"
#include "SparseLib++/gmres.h"
#include "SparseLib++/bicg.h"
#include "SparseLib++/cg.h"
#include "SparseLib++/bicgstab.h"
#include "common/tostring.h"
#include "engine/stiffnessmatrix.h"

void Solver::set_blocking(const std::vector<int> rb, 
			  const std::vector<int> cb) {
  rowblocks = rb;
  colblocks = cb;
}

void Solver::set_blocking(const FEMesh * mesh) {
  
  rowblocks = mesh->get_stiffness()->rowgroups(mesh->indepeqnmap);
  colblocks = mesh->get_stiffness()->colgroups(mesh->freedofmap);
}

GMRESSolver::~GMRESSolver()
{ 

}

bool GMRESSolver::solve(const CSRmatrix &A, const VECTOR_D &rhs, VECTOR_D &x) {
  
  
  //assume preconditioner has already been selected
  //  std::cerr << "GMRESSolver::solve: getting real preconditioner" << std::endl;
  Preconditioner * preconditioner  = pc->create_preconditioner
    (A, rowblocks, colblocks);
  //std::cerr << "GMRESSolver::solve: got preconditioner" << std::endl;
  MATRIX_D H(krylov_dimension+1, krylov_dimension);
  int krlv_dim = krylov_dimension;//will be modified by GMRES
  number_of_iterations = max_iterations;
  residual_norm = tolerance;
  //std::cerr << "GMRESSolver::solve: solving" << std::endl;
  status_ = GMRES(A, x, rhs, *preconditioner, H, krlv_dim, number_of_iterations, residual_norm);
  
  SolverMessager messenger(status_, number_of_iterations, residual_norm);
  _message = messenger._message();
  
  delete preconditioner;
  return !status_;
  
}

std::string GMRESSolver::message() const {
  return "GMRES" + _message;
}

bool BiCGSolver::solve(const CSRmatrix &A, const VECTOR_D &rhs, VECTOR_D &x) {
  //assume preconditioner has already been selected
  Preconditioner * preconditioner  = pc->create_preconditioner
    (A, rowblocks, colblocks);
  number_of_iterations = max_iterations;//will be modifed by method
  residual_norm = tolerance;
  status_ = BiCG(A, x, rhs, *preconditioner, number_of_iterations, residual_norm);

  SolverMessager messenger(status_, number_of_iterations, residual_norm);
  _message = messenger._message();

  delete preconditioner;
  return !status_;
  
}


BiCGSolver::~BiCGSolver(){
  //std::cerr <<"destroyed BiCG solver " << std::endl;
}

std::string BiCGSolver::message() const {
  return "BiCG" + _message;
}



bool CGSolver::solve(const CSRmatrix &A, const VECTOR_D &rhs, VECTOR_D &x) {
  //assume preconditioner has already been selected
  Preconditioner * preconditioner  = pc->create_preconditioner
    (A, rowblocks, colblocks);
  residual_norm= tolerance;//will be modifed by method
  number_of_iterations = max_iterations;
  status_ = CG(A, x, rhs, *preconditioner, number_of_iterations, residual_norm);

  SolverMessager messenger(status_, number_of_iterations, residual_norm);
  _message = messenger._message();

  delete preconditioner;
  return !status_;
}


CGSolver::~CGSolver(){
  //std::cerr <<"destroyed CG solver " << std::endl;
  

}

std::string CGSolver::message() const {
  return "CG" + _message;
}


bool BiCGStabSolver::solve(const CSRmatrix &A, const VECTOR_D &rhs, VECTOR_D &x) {
  //assume preconditioner has already been selected
  Preconditioner * preconditioner  = pc->create_preconditioner
    (A, rowblocks, colblocks);
  residual_norm = tolerance;//will be modifed by method
  number_of_iterations = max_iterations;
  status_ = BiCGSTAB(A, x, rhs, *preconditioner, number_of_iterations, residual_norm);

  SolverMessager messenger(status_, number_of_iterations, residual_norm);
  _message = messenger._message();

  delete preconditioner;
  return !status_;
}


BiCGStabSolver::~BiCGStabSolver(){
  //std::cerr <<"destroyed BiCGStab solver " << std::endl;
  

}

std::string BiCGStabSolver::message() const {
  return "BiCG Stabilized" + _message;
}


DirectSolver::~DirectSolver() {}

bool DirectSolver::solve(const CSRmatrix &A, const VECTOR_D &rhs, 
			 VECTOR_D &x) 
{

  SmallMatrix small_m(A.nrows(), A.ncols());
  small_m.clear();

  cProgressBar *cpb = getcProgressBar();

  for(int i=0; i<A.nrows(); i++) 
    for(int j=0; j<A.ncols(); j++)
      small_m(i,j) = A(i,j);  

  SmallMatrix small_r(rhs.size(),1);
  for(int i=0; i<rhs.size(); i++)
    small_r(i,0) = rhs(i);

  int rcode = small_m.solve(small_r);
  
  double r=0.0;
  if(rcode==0) {
    for(int i=0; i<rhs.size(); i++)
      x[i] = small_r(i,0);
    status_ = true;
    cpb->set_success();
    
    // Compute square of residual.
    double *resid = new double[A.nrows()];
    for(int i=0;i<A.nrows();i++) {
      resid[i]=rhs(i);
      for(int j=0;j<A.ncols();j++) 
	resid[i]-=A(i,j)*x[j];
      r += resid[i]*resid[i];
    }
    delete [] resid;
    
  }
  else {
    status_ = false;
    cpb->set_failure();
  }


  SolverMessager messenger((status_==true?0:1), 1, sqrt(r));
  _message = messenger._message();

  return status_;
}


std::string DirectSolver::message() const {
  return "Direct" + _message;
}



SolverMessager::SolverMessager(const int state, const int num_iter, const double res){
  switch (state){
    
    case 0:
      message = " converged!\n";
      break;
    case 1:
      message = " FAILED to converge!\n";
      break;
    
    default:
      message = " Unknown state";
      break;
  };
  std::string solver_info = "residual = " + to_string(res) + " number of completed iterations = " + to_string(num_iter);
  message += solver_info;

}

// The message is also the "repr" -- this can include some state data,
// and is potentially confusing, but it's mainly for type-identifying
// and debugging, so hopefully it won't confuse actual users.
// Function uses only public data, and so is nobody's friend.
std::ostream & operator<<(std::ostream &os, const Solver &s) {
  os << s.message();
  return os;
}

/*
bool linear_solver(Solver * solver, FEMesh * mesh){


  //at this point the solver and the preconditioner are already selected
  // at the UI and passed as arguments into the solver

  
  CSRmatrix A = mesh->Amatrix();

  
  if(A.nrows() != A.ncols())
    throw ErrSetupError("Stiffness matrix is not square! nrows="
			+ to_string(A.nrows()) + ", ncols="
			+ to_string(A.ncols()));
  
  
  CSRmatrix C = mesh->Cmatrix();

  int n = A.nrows();
  int m = C.ncols();
  VECTOR_D x(n, 0.0);
  VECTOR_D rhs(n, 0.0);
  VECTOR_D xp(m, 0.0);
  RWLock * rwlock = mesh->rwlock;
  rwlock->read_acquire();
  mesh->get_unknowns(x, mesh->freedofmap);

  mesh->get_rhs(rhs, mesh->indepeqnmap);
  mesh->get_unknowns(xp, mesh->fixeddofmap);
  rwlock->read_release();

  rhs -= C*xp;
   
  A *= -1.0;  // Make the matrix +ve-definite for preconditioners.
  rhs *= -1.0;
  
  //TO THIS POINT, ALL SOLVERS SHOULD DO THE SAME
  

  //THIS SECTION IS FOR DEBUGGING
//   int n_debug = 5;
//   SparseLinkMatrix<double> A_slm(n_debug,n_debug);
//   A_slm(0,0) = 5.0;
//   A_slm(1,1) = 5.0;
//   A_slm(1,0) = 1.0;
//   A_slm(0,1) = 1.0;
//   A_slm(2,2) = 10.0e10;
//   A_slm(3,3) = 10.0e10;
//   A_slm(4,4) = 10.0e10;
  
//   SPARSE_MATRIX_D A_temp(A_slm);
//   //std::cerr << "A_temp at " << &A_temp << std::endl;
//   CSRmatrix A_debug (A_temp);
  
//   VECTOR_D rhs_debug(n_debug, 1.0);
//   VECTOR_D x_debug(n_debug, 0.0);
//   //Assume solver and preconditioner have been already selected
//   //Assume parameters have been set
//   //cerr << "x = \n" << x_debug <<endl;
//   solver->precompute(A_debug);
//   cerr <<"A = \n" << A_debug << endl;
//   solver->solve(A_debug, rhs_debug, x_debug);
//   cerr << "x = \n" << x_debug <<endl;
  
//   solver->postcompute();
  //SECTION IS FOR DEBUGGING ENDS HERE

  solver->set_blocking(mesh->get_stiffness()->rowgroups(mesh->indepeqnmap),
		       mesh->get_stiffness()->colgroups(mesh->freedofmap));
  solver->precompute(A);
  //cerr <<"A = \n" << A << endl;
  bool status = solver->solve(A, rhs, x);
  //cerr << "x = \n" << x <<endl;
  // std::cerr << "rhs = \n" << rhs << std::endl;
  // std::cerr << solver->message() << std::endl;
  solver->postcompute(); // Call possibly-null post-compute hook.

  rwlock->write_acquire();

//   std::cerr << "Setting unknonwns -- free DOF map." << std::endl;
//   for(int i=0; i<freedofmap.size(); i++) {
//     std::cerr << i << " : " << freedofmap[i] << std::endl;
//   }
  
  mesh->set_unknowns(x, mesh->freedofmap);
  
  //   std::cerr << "x on exit from the solver." << std::endl;
  //   for(int i=0; i<x.size(); i++) {
  //     std::cerr << i << " : " << x[i] << std::endl;
  //   }
  
  CSRmatrix Cp = mesh->Cpmatrix();
  CSRmatrix B = mesh->Bmatrix();
  
  VECTOR_D rhsp = Cp*x+B*xp;
  mesh->set_rhs(rhsp, mesh->depeqnmap); // <-- not modified to reflect local_rhs.
  rwlock->write_release();

  return status;

}
*/
