# -*- python -*-
# $RCSfile: oofimage_.py,v $
# $Revision: 1.24 $
# $Author: redwing $
# $Date: 2004/06/21 16:47:17 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 


from SWIG.common import ooferror
from common import debug
from common import parallel_enable
import common.microstructure
from common import primitives
from common.IO import whoville
import os


class ImageContext(whoville.WhoDoUndo):
    def __init__(self, name, classname, oofimage):
        whoville.WhoDoUndo.__init__(self, name, 'Image', oofimage)
        self.microstructure = None

    def addMicrostructure(self, ms):    # called by Microstructure.addImage()
        self.getObject().setMicrostructure(ms)      # sets C++ CMicrostructure
        self.microstructure = ms
    def removeMicrostructure(self, ms): # called by Microstructure.removeImage()
        self.getObject().removeMicrostructure()
        self.microstructure = None
    def getMicrostructure(self):
        return self.microstructure
    def getSelectionContext(self):
        # Returns the object that holds the pixel selection, not the
        # selection itself.
        if self.microstructure is not None:
            return self.microstructure.pixelselection
    def writeImage(self, datafile):
        from image.IO import imageIO    # delayed import to avoid loops
        self.begin_reading()
        imageIO.writeImage(datafile, self)
        self.end_reading()
        
    def parallelize(self):
        global imageContexts
        if parallel_enable.query() == 1:
            try:
                from image.IO import oofimageIPC
                from SWIG.common import mpitools
                if mpitools.rank() == 0:
                    oofimageIPC.ipcoofimagemenu.Initialize(image =imageContexts.getPath(self))
            except ImportError:
                pass
        return

    def parallelPushModification(self, obj=None):
        ## assume top-most image is going to be pushed to the back-end
        global imageContexts
        if parallel_enable.query() == 1:
            try:
                from image.IO import oofimageIPC
                from SWIG.common import mpitools
                if mpitools.rank() == 0:
                    try:
                        oofimageIPC.ipcoofimagemenu.Push(image =imageContexts.getPath(self))
                    except KeyError:
                        pass
            except ImportError:
                pass
        return
    def parallelUndoModification(self):
        global imageContexts
        if parallel_enable.query() == 1:
            try:
                from image.IO import oofimageIPC
                from SWIG.common import mpitools
                if mpitools.rank() == 0:
                    oofimageIPC.ipcoofimagemenu.Undo(image =imageContexts.getPath(self))
            except ImportError:
                pass
        return
    def parallelRedoModification(self):
        global imageContexts
        if parallel_enable.query() == 1:
            try:
                from image.IO import oofimageIPC
                from SWIG.common import mpitools
                if mpitools.rank() == 0:
                    oofimageIPC.ipcoofimagemenu.Redo(image =imageContexts.getPath(self))
            except ImportError:
                pass
        return

    
## The imageContexts is a WhoDoUndoClass that hosts
## a set of instances of the ImageContext class, which
## which are subclasses of the WhoDoUndo class.
## Each ImageContext object contains an actual OOFImage object.
## Please also see the note on microsctructure.microStructures
imageContexts = whoville.WhoDoUndoClass(
    'Image',
    100,
    parentClass=common.microstructure.microStructures,
    instanceClass=ImageContext,
    proxyClasses=['<topmost>'])

def readImage(filename, **kwargs):
    # 'height' and 'width' keywords are optional.  If neither is
    # given, the height and width are the image size in pixels.  If
    # just one is given, the other is determined from the aspect
    # ratio.  If both are given, the image may be distorted.
    basename = os.path.basename(filename)
    image = OOFImage(basename, filename)
    # Set physical size of image
    pixelsize = image.sizeInPixels()    # iPoint
    given_height = kwargs.has_key('height')
    given_width = kwargs.has_key('width')
    if not (given_height or given_width):
        width = float(pixelsize.x)
        height = float(pixelsize.y)
    elif given_width and given_height:
        width = float(kwargs['width'])
        height = float(kwargs['height'])
    else:
        aspect = float(pixelsize.x)/pixelsize.y
        if given_width:
            width = float(kwargs['width'])
            height = width/aspect
        elif given_height:
            height = float(kwargs['height'])
            width = height*aspect
    image.setSize(primitives.iPoint(width, height))

    return image

def _getTimeStamp_optional_args(self, *args):
    return self._getTimeStamp()
OOFImagePtr.getTimeStamp = _getTimeStamp_optional_args

##def _getMicrostructure(self):
##    cms = self.getCMicrostructure()
##    return common.microstructure.microStructures[cms.name()]
##OOFImagePtr.getMicrostructure = _getMicrostructure

def getImage(imagename):
    ## returns the actual image
    ## imagename = microstructurename:image_name OR imagename = [microstructurename, image_name]
    return imageContexts[imagename].getObject()

def copyImage(imagecontext, name):
    imclone = imagecontext.getObject().clone(name)
    imclone.setMicrostructure(self.microstructure)
    return imclone

def pushModification(imagename, image):
    imageContexts[imagename].pushModification(image)

def undoModification(imagename):
    imageContexts[imagename].undoModification()

def redoModification(imagename):
    imageContexts[imagename].redoModification()

def undoable(imagename):
    return imageContexts[imagename].undoable()

def redoable(imagename):
    return imageContexts[imagename].redoable()


ooferror.registerErrorClass(ImageMagickError)
